open Args
open Message
open Topology
open Message
open Cops_api
open MoveChooser
open Planner
open Info
open Fw

(* Le seuil au deca duquel on reprend notre plan et pas le vote. *)
let delta_vote_score = 0.
(* Le seuil a partir duquel une information est pertinente quant a 
   la presence du voleur. *)
let alert_threshold = 0
(* Si le voleur est a trop d'endroit, on ne sait plus ou il est. *)
let too_many_alerts = 25
let long_term_changing_threshold = 2.

let my_name = "crimee_cop"

type crimee_mode =
    Cover 
  | Hunt

type user_infos = 
    {
      (*   *) should_inform            : bool;   
      (*   *) last_positions           : nod array;
      mutable planner_state            : Planner.state option;
      mutable fw_state                 : Fw.fw option;
      mutable long_term_covering_plan  : plan list;
      mutable long_term_covering_power : float;
    }

let user_infos_init =
  {
    should_inform               = true;
    last_positions              = [||];
    fw_state			= None;
    planner_state               = None;
    long_term_covering_power    = 0.;
    long_term_covering_plan     = [];
  }

let get_planner_state topo draw user_infos = 
  match user_infos.planner_state with
    | None -> 
	let st = Planner.init topo draw in
	  user_infos.planner_state <- Some st;
	  st
    | Some x -> x

let get_fw_state topo user_infos = 
  match user_infos.fw_state with
    | None -> 
	let st = Fw.init_fw topo in 
	  user_infos.fw_state <- Some st;
	  st
    | Some x -> x
    
(* [update_profile w user_infos] looks at the effective moves of cops
   to determine if there are lyers or not. *)
let update_profile world user_infos =
  user_infos

let static_eval_plan env w user_infos plan =
  0.

let determine_hunt_plan env world user_infos = 
  ([], 0.)

let min_distance fw_state topo node node' ptype = 
  let (length, result) = 
    shortest_path_next_step fw_state ptype node node' 
  in
    match result with
      | None -> 
	  (Cop_car, max_int)
      | Some result ->
	  (result.fns_ptype_final, Misc.F_INT.int_of_fint length)

(* let closest_bank env world bank =  *)
(*   let banks = env.env_topo.t_banks in *)
(*   let node = node_of_player env.env_topo cop in *)
(*   let _, bank_node =  *)
(*     try NS.fold  *)
(*         (fun bank_node (min, min_node) ->  *)
(* 	      let d = min_distance env.env_topo node bank_node cop.pl_ptype  *)
(* 	      in *)
(* 		if d < min then *)
(* 		  (d, bank_node) *)
(* 		else *)
(* 		  (min, min_node)) banks (max_int, NS.choose banks) *)
(*     with Not_found -> failwith "No banks (!)." *)
(*   in  *)
(*     bank_node *)

let nearest_player fw_state env bank_node cops =
  List.fold_left (fun (min_d, min_ptype, min_cop) cop -> 
	    if is_cop cop then 
	      let player_node = node_of_player env.env_topo cop in
	      let ptype, d = 
		min_distance fw_state env.env_topo player_node.nod_num
		  bank_node.nod_num cop.pl_ptype
	      in
	      let d2 = d * d in
		if min_d > d2 then
		  (d2, ptype, cop)
		else 
		  (min_d, min_ptype, min_cop)
	    else (min_d, min_ptype, min_cop)) (max_int, Cop_car, List.hd cops)
    cops

(* We are computing the covering of cops in the world so as to 
   protect the banks. [between 0..1]
   product (j, 1 / (1 + min_i (d(bank_i, cop_i)^2)))
   We are not taking into acount the money in the banks and 
   We should also promote dynamic.
*)
let eval_covering_power fw_state env cops =
  NS.fold 
    (fun bank acu -> 
       let d, _, _ = nearest_player fw_state env bank cops in
	 1. /. (1. +. (float_of_int d)) *. acu)
    env.env_topo.t_banks
    1.
  
(* Ceci est une heuristique pour maximiser la fonction precedente. 
   On fait se rapprocher les policiers de la banque la plus proche qui
   n'est pas deja affectee (sinon la deuxieme etc ...).
*)    
let determine_covering_plan (env : env) (world : world) user_infos = 
  let fw_state = get_fw_state env.env_topo user_infos in

  (* Determine le plan sur du "long" terme. Affectation d'une banque
     a chaque policier. *)
  let long_term = 
    let rec assign bank_node (remaining_cops, result) =
	if remaining_cops = [] then
	  (remaining_cops, result)
	else 
	  let (_, _, cop) as nearest = 
	    nearest_player fw_state env bank_node remaining_cops
      in
      let cops = List.filter (fun c -> c <> cop) remaining_cops in
	(cops, (bank_node, nearest) :: result)
    in
  let _, cops_of_banks = 
    NS.fold assign env.env_topo.t_banks (List.filter is_cop world.wor_pls, []) 
  in
    List.fold_left (fun plan (bank_node, (d, ptype, cop)) -> 
	      {
		plan_bot   = cop.pl_bot;
		plan_loc   = bank_node.nod_loc;
		plan_ptype = ptype;
		plan_world = d;
	      } :: plan) [] cops_of_banks
  in
  let new_cops = List.map (fun plan -> 
			   {
			     pl_bot   = plan.plan_bot;
			     pl_loc   = plan.plan_loc;
			     pl_ptype = plan.plan_ptype
			   }) long_term
  in
  let new_covering_power = eval_covering_power fw_state env new_cops in
    if new_covering_power /. user_infos.long_term_covering_power >
      long_term_changing_threshold then (
	user_infos.long_term_covering_plan <- long_term;
	user_infos.long_term_covering_power <- new_covering_power);
    (user_infos.long_term_covering_plan, user_infos.long_term_covering_power)

let determine_plan = function
  | Hunt  -> determine_hunt_plan 
  | Cover -> determine_covering_plan

let inv_compare_float (r1 : float) (r2 : float) = 
  if r1 -. r2 < min_float then 1
  else if r1 = r2 then 0 
  else -1
    
let which_mode env world potentials connect =
  let card_potentials = ref 0 in
    Array.iter (fun x -> if x > alert_threshold then incr card_potentials) 
      potentials;
    if !card_potentials > too_many_alerts then
      Cover
    else 
      Hunt

let crimee_cop me env world user_infos =
  let my_name = env.env_wsk.wsk_name in
  let info = env.Cops_api.env_info
  and topo = env.Cops_api.env_topo in
  let planner_state = get_planner_state topo env.env_drawer user_infos in
  let fw_state = get_fw_state topo user_infos in
  let possible = Info.get_possible info
  and certain = Info.get_certain info 
  and node, ptype = Info.get_me info in
  let connect = Connected.decompose topo possible in
  let coefs = { static = 0; robber = 1; lastv = 0; repulse = 1 } in
  let potentials = 
    Array.map (fun node -> potential_node planner_state connect 
	       certain possible world.wor_wor ptype node coefs)
      topo.t_nods
  in
    Draw.display_node_sets env.env_drawer
      (Info.get_possible info)
      (Info.get_certain info);
    
  Array.iteri
    (fun i n -> Draw.draw_label env.env_drawer topo.t_nods.(i) (string_of_int n))
    potentials;

  let mode = which_mode env world potentials connect in
  let mode = Cover in

    (* Look at the previous cop moves to enrich their profiles. *)
  let user_infos = update_profile world user_infos in

    (* We are sharing information except if we think it is not a good idea. *)
  let informs = 
    let our_information = 
      if user_infos.should_inform then 
	Cop_informer.compute_inform info
      else 
	[]
    in
      cop_inform me env our_information 
  in

    (* Use alpha-beta to determine a good plan. *)
    let plan, our_plan_score = determine_plan mode env world user_infos in
      List.iter dump_plan plan;
    (* Vote using our evaluation function. *)
    let plans = cop_plan me plan in
    let other_plans = List.filter (fun (n, p) -> n <> my_name) plans in
    let plans_scores = 
      List.map (fun (n, p) -> n, 
		static_eval_plan env world user_infos p) other_plans 
    in
    let sorted_plans = 
      List.sort (fun (_, r1) (_, r2) -> inv_compare_float r1 r2) plans_scores 
    in
    let all_scored_plans = (my_name, our_plan_score) :: sorted_plans in
    let vote = cop_vote me (fst (List.split all_scored_plans)) in

      (* On choisit le plan vote si il est raisonnable. *)
    let chosen_plan = 
      match vote with
	| Winner winner -> 
	    Printf.eprintf "Winner %s%!\n" winner;
	    (* Is this plan good enough ? *)
	    let winner_score = List.assoc winner all_scored_plans in
	      if our_plan_score -. winner_score < delta_vote_score then
		List.assoc winner plans
	      else 
		plan
	| Nowinner -> 
	    Printf.eprintf "NoWinner\n%!";
	    plan
    in
      (* On suit le plan choisi. *)
    let chosen_move = 
      find_move_according_to_plan fw_state env world user_infos chosen_plan plan
    in
      Cops_api.dump_mov chosen_move;
      (chosen_move, user_infos)

let () = 
  register_cop my_name
    { cops_init = (fun _ env -> 
		     Voting.init env.env_wsk; 
		     user_infos_init);
      cops_job  = crimee_cop;
      cops_ptype = Cop_car;
    }

