(* $Id: damien_cop2.ml,v 1.16.2.26 2005/07/10 13:54:00 frisch Exp $ *)

(* Attention: on suppose ici que les plans evalus sont valides (sinon ca plante) *)

(* TODO: preprocessing sur chaque item de plan avant produit cartsien *)

open Info
open Message
open Printf
open Topology
open Cops_api

type state = {
  mutable ask_for_corruption: bool;
  voter: Voting.t;
  r_voter: Voting.t;
}

let rec cross_product l =
  match l with
  | [] -> [ [] ]
  | []::t -> []
  | (h::t)::rest ->
      List.map (fun x -> h :: x) (cross_product rest)
      @ cross_product (t::rest)
;;

(* Construire tous les plans possibles *)
let possible_plans info : (int * (Topology.NS.elt * Message.ptype)) list list =
  let topo = topo info in
(*  let cops = map_cops info (fun nam i nod typ -> (nam, (nod, typ))) in *)
(*  let plan_item name (nod, typ) = {
    plan_bot = name;
    plan_loc = nod.nod_loc;
    plan_ptype = typ;
    plan_world = current_world info + 1;
  } in *)
  let brains = Info.brains info in
  let moves1 _ i nod typ =
    let dests = NS.elements (possible_dest topo nod typ) in
    List.map (fun x -> (i,(x,typ))) dests
  in
  let moves = map_cops info moves1 in
  cross_product moves
;;


(* Encourager les 2 flics les plus proches  tre  distance 2 ou 3 *)
let position_bonus topo nod (l : (int * (Topology.NS.elt * Message.ptype)) list) =
  let dist (_,(nod',_)) = dist_f topo nod nod' in
  let score = function
    | 0 -> 0
    | 1 -> 3
    | 2|3 -> 6
    | 4 -> 3
    | 5 -> 2
    | _ -> 1
  in
  match List.sort compare (List.map dist l) with
  | _::d1::d2::_ -> score d1 + score d2
      (* Le premier element est 0 (le flic lui-meme) *)
  | _ -> 0 (* assert false *)
;;

let evaluate info (p : (int * (Topology.NS.elt * Message.ptype)) list) =
  let topo = topo info in


  (* TODO: faire ce calcul une fois pour toute *)

  let next = extend topo (get_possible info) in
  let possible_robber = ref next in
  (* ensemble des positions ou le voleur pourra tre au coup d'aprs *)

  let brains = Info.brains info in
  let players = Array.create (Array.length topo.cops_name) [] in

  let positions = ref 0 in
  let f (i,((nod,typ) as m)) =
    let controler = Brains.controller_i brains i in
    players.(controler) <- m :: players.(controler);
    possible_robber := NS.remove nod !possible_robber;
    positions := !positions + position_bonus topo nod p;
  in
  List.iter f p;

  let unhide = ref NS.empty in
  (* ensemble des positions couvertes par les odeurs *)
  Array.iter 
    (fun cops ->
       if cops <> [] then unhide := NS.union !unhide (player_nosmell topo cops)
    ) players;
  let hidden_robber = NS.diff next !unhide in
  (* ensemble des positions ou le voleur pourra tre au coup d'aprs
     s'il n'est pas sniff (au coup d'aprs) *)


  let distances = 0 in
(*
  let distances = NS.fold (fun e d -> d + dist_to_set topo e !cop_set)
                  !possible_robber 0
  in
*)
  let distance2 = ref 0 in
  let f (_,(nod,typ)) =
    distance2 := !distance2 + dist_to_set_typ topo nod !possible_robber typ
  in
  List.iter f p;


(* Critres lexicographiques:
   - diminuer ensemble cachette
   - diminuer ensemble possible
   - sinon, se rapprocher
   - et disperser les flics *)
(* TODO: critres secondaires en Lazy *)

(* TODO: elager construction du produit cartesien pour ne pas
   s'eloigner de (extend?) hidden_robber *)
  (NS.cardinal hidden_robber,
   NS.cardinal !possible_robber,
   distances,
   !distance2,
   !positions)
;;

let print_score (hr,pr1,_,d,p) =
  Printf.eprintf "(%i,%i,%i,%i)" hr pr1 d p

let rec dump_first_plans n p = 
  match n,p with
    | 0,_ | _,[] -> ()
    | n,hd::tl ->
	print_score (snd hd);
	dump_first_plans (pred n) tl

let compare_plans (_, (hr1, pr1, d1, dd1, p1)) (_, (hr2, pr2, d2, dd2, p2)) =
  if hr1 < hr2 then -1
  else if hr1 > hr2 then 1
  else if pr1 < pr2 then -1
  else if pr1 > pr2 then 1
(*
  else if d1 < d2 then -1
  else if d1 > d2 then 1
*)
  else begin
    let score1 = 10*dd1 - p1 in
    let score2 = 10*dd2 - p2 in
    (* La dispersion est moins importante que le rapprochement du voleur,
       mais pas trop quand mme (rien ne sert de se marcher sur les pieds) *)
    if score1 < score2 then -1
    else if score1 > score2 then 1
    else 0
  end;
;;

let get_my_moves mycops info p =
  let t = topo info in
  let m = ref [] in
  List.fold_left 
    (fun accu x ->
       let i = Topology.cop_number t x.plan_bot in
       if List.mem i mycops 
       then (i,(x.plan_ptype, get_node t x.plan_loc))::accu
       else accu
    ) [] p


let bottom_score = ref (None: (int * int * int * int * int) option)

(* Update bottom_score with the 80-percentile score
   (the score that is >= to the score of 80% of the possible moves) *)

let update_bottom_score evplans =
  try
    bottom_score := Some (snd (List.nth evplans (List.length evplans / 5)))
  with Failure _ ->
    bottom_score := None

let make_pursuit_plan info =
  let plans = possible_plans info in
  let evplans = List.map (fun x -> (x, evaluate info x)) plans in
  let plans = List.sort compare_plans evplans in
  dump_first_plans 5 plans;
  Printf.eprintf "\n"; flush stderr;
  update_bottom_score plans;
  fst (List.hd plans)
    (* TODO: extraire le minimum en linaire *)
;;


(* Plan de rechange quand on ne sait pas trop ou le voleur se trouve:
   on essaie de sniffer un peu partout. *)

let never_sniffed = ref NS.empty;;

let update_sniffed env =
  let topo = env.env_topo in
  let info = env.env_info in
  let brains = Info.brains info in
  let players = Array.create (Array.length topo.cops_name) [] in
  iter_cops env.env_info (fun _ i nod typ ->
    let controler = Brains.controller_i brains i in
    players.(controler) <- (nod,typ) :: players.(controler);
  );
  let unhide = ref NS.empty in
  Array.iter 
    (fun cops ->
       if cops <> [] then unhide := NS.union !unhide (player_nosmell topo cops)
    ) players;
  never_sniffed := NS.diff !never_sniffed !unhide

let make_explore_plan corrupted info =
  bottom_score := None;
  let topo = topo info in
(*  update_sniffed topo info; *)

  let set = 
    if not corrupted then
      NS.inter !never_sniffed (extend topo (get_possible info))
    else !never_sniffed in
  let set =
    if NS.is_empty set then (never_sniffed := topo.all_nodes; get_possible info)
    else set in

(*  Draw.display_node_sets (drawer info)
    ~title:"never_sniffed" set NS.empty; *)

  map_cops info (fun name i nod typ ->
		   (* Choisir au hasard une manire
		      de se rapprocher de l'ensemble *)
		   (* On pourrait mettre comme premier critre
		      de diminuer cet ensemble *)
    let dests = NS.elements (possible_dest topo nod typ) in
    let best = ref [nod] in
    let dist n = dist_to_set_typ topo n set typ in
    let best_dist = ref (dist nod) in
    List.iter (fun n ->
      let d = dist n in
(*      Draw.draw_label (drawer info) n (string_of_int d); *)
      if d < !best_dist then begin
        best_dist := d;
        best := [n];
      end else if d = !best_dist then best := n :: !best)
      dests;
    let l = List.length !best in
    let choice = List.nth !best (Random.int l) in
    (i,(choice,typ))
  )
;;

exception Timeout

(* TODO: Si never_sniffed devient vide, le remettre  all_nodes *)

let swap l = List.map (fun (i,(x,y)) -> (i,(y,x))) l

let make_plan corrupted mycops env =
  let info = env.env_info and topo = env.env_topo in
  Printf.eprintf "world %i: possible nodes %i\n" (current_world info) (NS.cardinal (get_possible info));
  flush stderr;
  bottom_score := None;
  let the_plan =
    if corrupted || (NS.cardinal (get_possible info)) > 60 then begin
      (if corrupted then
	 match (Info.get_robber info) with
	   | Some b ->
	       never_sniffed := 
		 NS.diff !never_sniffed 
		   (extend topo (extend topo (NS.singleton b)))
	   | None -> ());
(*
      eprintf "plan: explore\n"; flush stderr;
*)
      make_explore_plan corrupted info
    end else begin
      Utils.timeout 4 "Failing on pursuit plan, back to explore" ()
	~normal:(fun () ->
		   (*	eprintf "plan: pursuit\n"; flush stderr; *)
		   let r = make_pursuit_plan info in
		   never_sniffed := topo.all_nodes;
		   r)
	~timedout:(fun () ->
		     make_explore_plan corrupted info)
    end
  in
  let the_plan = 
    if corrupted then swap (Planify.degrade_plan info (swap the_plan))
    else the_plan in

  let the_plan = Planify.mk_plans info the_plan in


  (get_my_moves mycops info the_plan, the_plan)
;;

 
  

let type_to_string = function
  | Cop_car -> "car"
  | Cop_foot -> "foot"
  | Robber -> "robber"
;;

let debug_plan name plan =
  let f pi =
    if pi.plan_bot = name then begin
      eprintf "plan: %s -> %s on %s at %d\n"
              pi.plan_bot
              pi.plan_loc
              (type_to_string pi.plan_ptype)
              pi.plan_world;
      flush stderr;
    end
  in
  List.iter f plan;
;;


let exchange_informs me env world = 
  let infos = Cop_informer.compute_inform env.env_info world in
  cop_inform me env infos

let exchange_plans me voter topo plan = 
  let plans = cop_plan me plan in
  let vote = Voting.vote voter topo.name (List.map fst plans) in
  let vote_res = cop_vote me vote in
  Voting.count voter vote_res;
  plans,vote_res

let change_transport env = 
  env.env_info <- 
    change_cop env.env_info (Planify.change_cops_move_type env.env_info)

let trace env = 
  let poss = get_possible env.env_info in
  Draw.display_node_sets env.env_drawer 
    ~title:(string_of_int (NS.cardinal poss)) poss NS.empty

let validate_voted_moves info the_moves =
  match !bottom_score with
  | None -> true (* we have no scores, so choose voted move *)
  | Some bottom_score ->
      let the_score =
        evaluate info
          (List.map (fun (x, (ptype, node)) -> (x, (node, ptype))) 
                    the_moves) in
        compare_plans ((), the_score) ((), bottom_score) < 0

let choose_move env w my_moves plans vote_res =
  match vote_res with
    | Nowinner -> my_moves
    | Winner name ->
	(* Suivre le plan du leader s'il est legal pour moi *)
	(* TODO: accepter de changer de mode de transport *)
	let my_cops = Info.get_my_cops env.env_info in
	try
          let the_plan = snd (List.find (fun (x, _) -> x = name) plans) in
          let the_moves = get_my_moves my_cops env.env_info the_plan in
	  if not (Topology.is_legal env.env_topo w the_moves) then raise Not_found;
          if validate_voted_moves env.env_info the_moves then begin
            Printf.eprintf "Follow Winner plan: %s\n" name;
            the_moves
          end else begin
            Printf.eprintf "Discard Winner plan: %s (not good enough)\n" name;
            my_moves
          end
	with Not_found -> my_moves

let want_corrupt env =
  (* Est-ce que je veux etre corrompu ? *)
  let info = env.env_info in
  let poss = get_possible env.env_info in
  (Info.current_world info >= 180) && (NS.cardinal poss >= 30) ||
    (Info.current_world info >= 50) && (NS.cardinal poss >= 30) &&
    (Info.robbed info >= 2500)

let damien_cop2 me env w state =
  update_sniffed env;
  change_transport env;

  let (_ : (string * inf list) list) = exchange_informs me env w in
  trace env;

  let my_moves,plan = make_plan false (Info.get_my_cops env.env_info) env in
  let plans,vote_res = exchange_plans me state.voter env.env_topo plan in
  let move = choose_move env w my_moves plans vote_res in
  let move = Planify.improve_plan env.env_info move in

  if want_corrupt env then state.ask_for_corruption <- true;

  (if state.ask_for_corruption then Turncoat else Straight),
  move, 
  Info.get_ripoux env.Cops_api.env_info,
  state

let corrupted me env w state =
  Printf.eprintf "I (=%s) am corrupted !!!\n" env.env_topo.name;
  update_sniffed env;
  change_transport env;

  let (_ : (string * inf list) list) = exchange_informs me env w in
  trace env;

  let my_moves,plan = make_plan true (Info.get_my_cops env.env_info) env in
  let plans,vote_res = exchange_plans me state.voter env.env_topo plan in
  let move = choose_move env w my_moves plans vote_res in
  let move = Planify.degrade_plan env.env_info move in

  move, 
  state


let corrupted_r me env w state =
  let informs = cop_r_inform me env [] in
  let plans,vote_res = exchange_plans me state.r_voter env.env_topo [] in
  state

  

let () =
  register_cop "damien_cop2"
    { cops_init = (fun _ env -> 
		     { voter = Voting.init env.env_wsk;
		       r_voter = Voting.init env.env_wsk;
		       ask_for_corruption = false });
      cops_clean_job  = damien_cop2;
      cops_dirty_job = corrupted;
      cops_dirty_r_job = corrupted_r;
      cops_ptype = Cop_foot;
    }

let () =
  register_cop "damien_cop2_corrupt"
    { cops_init = (fun _ env -> 
		     { voter = Voting.init env.env_wsk;
		       r_voter = Voting.init env.env_wsk;
		       ask_for_corruption = false });
      cops_clean_job  = damien_cop2;
      cops_dirty_job = corrupted;
      cops_dirty_r_job = corrupted_r;
      cops_ptype = Cop_foot;
    }
