(* $Id: message.ml,v 1.15.4.4 2005/07/10 00:08:55 lefessan Exp $ *)

open Printf;;

type ptype =
  | Cop_foot
  | Cop_car
  | Robber
;;

type node_tag =
  | Hq
  | Bank
  | Robber_start
  | Ordinary
;;

type edge_type =
  | Car
  | Foot
;;

(* ---------------------- *)

let nnodes = ref 0
type nod = {
    nod_num : int;
    nod_loc : string;
    nod_tag : node_tag;
    nod_x : int;
    nod_y : int;
  }

type edg = {
  edg_src : string;
  edg_dst : string;
  edg_type : edge_type;
}

type bv = {
  bv_loc : string;
  bv_value : int;
}

type ev = {
  ev_loc : string;
  ev_world : int;
}

type pl = {
  pl_bot : string;
  pl_loc : string;
  pl_ptype : ptype;
}

(* ---------------------- *)

type register = {
  reg_bot : string;
  reg_ptype : ptype;
}

type world_skeleton = {
  wsk_name : string;
  wsk_robber : string;
  wsk_cops : string list;
  wsk_nods : nod list;
  wsk_edgs : edg list;
}

type sc = {
  sc_brain: string;
  sc_cop: string;
}

type fac = {
  fac_accusing: string;
  fac_accused: string;
  fac_wor: int;
}

type world = {
  wor_wor : int;
  wor_rbd: int;
  wor_dc: string list; (* known dirty cops *)
  wor_sc: sc list; (* cops taken over *)
  wor_fac: fac list; (* false accusations *)
  wor_bvs: bv list;
  wor_evs: ev list;
  wor_smell: int;
  wor_pls: pl list;
}

type offer = Turncoat | Straight

type move = {
  mov_loc : string;
  mov_ptype : ptype;
  mov_bot : string;
}

type moves = move list

type clean_cop_move = {
  mov_offer: offer;
  mov_moves: moves;
  mov_accuse: string list;
}

type dirty_cop_move = move list

type bribing = Bribe | Nobribe

type offered_cops =
  | NotTelling
  | Offered of string list

type push = {
  push_bot: string;
  push_loc: string;
}

type bribe_result =
  | Push of push
  | Choose of string list
  | Nothing

type robber_move = {
  mov_mov: move;
  mov_bribe: bribe_result;
}


type world_or_game_over =
  | World of world
  | Game_over
;;

type inf = {
  inf_bot : string;
  inf_loc : string;
  inf_ptype : ptype;
  inf_world : int;
  inf_certainty : int;
}

type inform = inf list;;

type plan = {
  plan_bot : string;
  plan_loc : string;
  plan_ptype : ptype;
  plan_world : int;
}

type vote = string (* Now a cop-brain, not a cop-bot. *)

type vote_res =
  | Winner of string
  | Nowinner
;;


let ptype_to_string p =
  match p with
  | Cop_car -> "cop-car"
  | Cop_foot -> "cop-foot"
  | Robber -> "robber"
;;

let register r =
  printf "reg: %s %s\n" r.reg_bot (ptype_to_string r.reg_ptype);
  flush stdout;
;;

let move m =
  printf "mov: %s %s %s\n" m.mov_loc (ptype_to_string m.mov_ptype) m.mov_bot;
  flush stdout;
;;

let moves m = 
  printf "mov\\\n";
  List.iter move m;
  printf "mov/\n";
  flush stdout

let clean_cop_move m =
  printf "cmov\\\n";
  (match m.mov_offer with
     | Turncoat -> printf "turncoat:\n"
     | Straight -> printf "straight-arrow:\n");
  moves m.mov_moves;
  printf "acc\\\n";
  List.iter (fun s -> printf "acc: %s\n" s) m.mov_accuse;
  printf "acc/\n";
  printf "cmov/\n";
  flush stdout

let dirty_cop_move = moves

let robber_move m =
  printf "rmov\\\n";
  move m.mov_mov;
  (match m.mov_bribe with
     | Push p -> printf "psh: %s %s\n" p.push_bot p.push_loc
     | Choose c -> 
	 printf "chc\\\n";
	 List.iter (fun n -> printf "chc: %s\n" n) c;
	 printf "chc/\n"
     | Nothing ->
	 printf "nobribe:\n");
  printf "rmov/\n";
  flush stdout
    
let bribing bribing = 
  begin
    match bribing with
    | Bribe -> printf "bribe:\n"
    | Nobribe -> printf "nobribe:\n"
  end;
  flush stdout
  
let vote vl =
  printf "vote\\\n";
  List.iter (fun s -> printf "vote: %s\n" s) vl;
  printf "vote/\n";
  flush stdout;
;;

let print_one_plan p =
  printf "plan: %s %s %s %d\n" p.plan_bot p.plan_loc
         (ptype_to_string p.plan_ptype) p.plan_world;
;;

let plan pl =
  printf "plan\\\n";
  List.iter print_one_plan pl;
  printf "plan/\n";
  flush stdout;
;;

let print_one_inf i =
  printf "inf: %s %s %s %d %d\n" i.inf_bot i.inf_loc
         (ptype_to_string i.inf_ptype) i.inf_world i.inf_certainty;
;;

let inform il =
  printf "inf\\\n";
  List.iter print_one_inf il;
  printf "inf/\n";
  flush stdout;
;;

let tag_to_string = function
  | Hq -> "hq"
  | Bank -> "bank"
  | Robber_start -> "robber-start"
  | Ordinary -> "ordinary"
;;


(* Default value for plan. *)
let default_plan =
  {
    plan_bot = "";
    plan_loc = "";
    plan_ptype = Cop_car;
    plan_world = min_int;
  }

(* [is_headquarter node] is true if the headquarter is here. *)
let is_headquarter node =
  node.nod_tag = Hq

(* [neg_ptype ptype] negate ptype for cops (car << = >> foot) *)
let neg_ptype = function
  | Cop_car  -> Cop_foot
  | Cop_foot -> Cop_car
  | Robber   -> Robber

let is_cop player =
  match player.pl_ptype with
    | Cop_car | Cop_foot -> true
    | Robber -> false

let string_of_plan plan =
  sprintf "plan: %s to %s by %s at %d\n"
    plan.plan_bot plan.plan_loc (ptype_to_string plan.plan_ptype)
    plan.plan_world

let dump_plan plan =
  eprintf "%s\n%!" (string_of_plan plan)

