(* $Id: robber.ml,v 1.17.2.5 2005/07/10 12:07:17 lefessan Exp $ *)


(* 

TODO list:

* Si on arrive a une banque en n mod 8 <> 0, on peut attendre qques tours
pour rentrer dans la banque uniquement en n mod 8 = 0.
* Si un flic reste a cote d'une banque, on ne s'interesse pas a cette banque.
* Ne pas aller systematiquement vers la banque la plus proche.
* Fixer les destinations une fois pour toute
*)

open Args
open Message
open Robber_types
open Robber_topology
  
(* [init world_skeleton] initialises the state with the current world
skeleton *)
  

  
let initialized = ref false

open Graphics
  
let current_robber = ref (None : robber option)
  
let (robbers : (string, robber * string) Hashtbl.t) = Hashtbl.create 13
let register_robber name help robber = 
  Hashtbl.add robbers name (robber, help)

let robber_automata rob =
  let t = Robber_topology.map_create () in
(*  let begin_time = Unix.gettimeofday () in *)
  register {
    reg_bot = our_robber_name;
    reg_ptype = Robber;
  };
(*  let registered_time = Unix.gettimeofday () in 
  Printf.fprintf stderr "Registered_time : %.2f s\n" (registered_time -. begin_time);
  flush stderr; *)
  let world_skeleton = Parser.world_skeleton Lexer.token lexbuf in
  Robber_topology.map_init t world_skeleton;
  Robber_draw.init t;
  
  (try rob.robber_init t with e -> 
        Printf.fprintf stderr "Exception %s in robber_init\n" 
          (Printexc2.to_string e));
(*  let init_time = Unix.gettimeofday () in
  Printf.fprintf stderr "Init time : %.2f s\n" (init_time -. begin_time);
  flush stderr; *)
  let rec iter () =
    let msg = Parser.world_or_game_over Lexer.token lexbuf in
    let begin_time = Unix.gettimeofday () in
    match msg with
      Game_over -> exit 0
    | World world ->
        let state = state_create t rob world in
        
        t.proposed_robber_bribe <- Nothing;
        Robber_misc.update_nodes t state rob;
        Robber_possibles.inform t state;

(* C -> S : inform-msg *)

        let inform = rob.robber_inform t state in
        Message.inform inform;
        
(* S -> C : from-msg-inform *)
        
        let from_inform = Parser.from_inform Lexer.token lexbuf in
        
        let plan = rob.robber_planning t state from_inform in
        
        Message.plan plan;
        
(* C -> S : plan-msg *)
        
(* S -> C : from-msg-plan *)
        
        let from_plan = Parser.from_plan Lexer.token lexbuf in
        let vote = rob.robber_voting t state from_plan in
        Message.vote vote;
        
(* C -> S : vote *)
        
(* S -> C : vote-res-msg *)
        let vote_res = Parser.vote_res Lexer.token lexbuf in
        let bribe = rob.robber_waiting t state vote_res in
        Message.bribing bribe;
(* C -> S : bribe-msg *)
        
(* S -> C : offered-cops *)
        let offered_cops = Parser.offered_cops Lexer.token lexbuf in
        rob.robber_moving t state offered_cops;
(* Check that the proposed robber node is reachable *)
        
        if not (List.mem t.proposed_robber_node 
              t.foot_edges.(t.robber_node)) then begin
            Printf.fprintf stderr "Proposed robber move is not OK\n";
            t.proposed_robber_node <- t.robber_node;
          end;
        
        Printf.fprintf stderr "Next position: %d\n" t.proposed_robber_node;
        (*
        let end_time = Unix.gettimeofday () in
        if end_time -. begin_time > 4.5 then begin
            Printf.fprintf stderr "***************************************\n";
            Printf.fprintf stderr "***************************************\n";
            Printf.fprintf stderr "***************************************\n";
            Printf.fprintf stderr "***************************************\n";
            Printf.fprintf stderr "****        TIMEOUT DANGER    *********\n";
            Printf.fprintf stderr "***************************************\n";
            Printf.fprintf stderr "***************************************\n";
            Printf.fprintf stderr "***************************************\n";
            Printf.fprintf stderr "***************************************\n";
            Printf.fprintf stderr "***************************************\n";
            exit 2;
          end;
        flush stderr;
*)        
        
        Robber_draw.wait t state;
        t.last_robber_node <- t.robber_node;
        t.robber_node <- t.proposed_robber_node;
        
        
        if t.last_robbery_time > 0 then begin
            t.last_robbery_time <- t.last_robbery_time - 1;
            if t.last_robbery_time = 0 then
              t.last_robbery <- -1
          end;
        let node = t.nodes.(t.robber_node) in
        begin
          match node.node_tag with
            Bank bank ->
              t.last_robbery_time <- 10;
              t.last_robbery <- bank.bank_node;
              t.robberies <- t.robberies + 1;
              t.stolen <- t.stolen + bank.bank_value
          | _ -> ()
        end;
        
        Message.robber_move {
          mov_mov = { 
            mov_loc = loc_of_node t t.robber_node;
            mov_ptype = Robber;
            mov_bot = t.our_name;
          };
          mov_bribe = t.proposed_robber_bribe;
        };
        flush stderr;
        iter ()
  in
  try
    iter ()
  with Parsing.Parse_error -> 
      Printf.fprintf stderr "Parse error on %s\n"  (Lexing.lexeme lexbuf);
      flush stderr;
      exit 2
