(* $Id: robber7.ml,v 1.2.4.2 2005/07/10 12:07:17 lefessan Exp $ *)


(* 

TODO list:

* Si on arrive a une banque en n mod 8 <> 0, on peut attendre qques tours
pour rentrer dans la banque uniquement en n mod 8 = 0.
* Si un flic reste a cote d'une banque, on ne s'interesse pas a cette banque.
* Ne pas aller systematiquement vers la banque la plus proche.
* Fixer les destinations une fois pour toute
*)

open Args
open Message
open Robber_types
open Robber_topology
open Robber

let copsum_threshold = 35
  
let compute_cops_proximity t w dist_threshold i = 
  let sum = ref 0 in
  for cop = 0 to t.ncops - 1 do
    match w.cops_ptypes.(cop) with
      Cop_car ->
        if t.car_dist.(w.cops_nodes.(cop)).(i) < dist_threshold then begin
(*            Printf.fprintf stderr "Car cop closed to %d : %d\n" 
              i t.car_dist.(w.cops_nodes.(cop)).(i); *)
            incr sum
          end
    | Cop_foot ->
        if t.foot_dist.(w.cops_nodes.(cop)).(i) < dist_threshold then begin
(*            Printf.fprintf stderr "Foot cop closed to %d : %d\n" 
              i t.foot_dist.(w.cops_nodes.(cop)).(i); *)
            incr sum
          end
    | Robber -> assert false
  done;
  !sum
  
  
let compute_node_value t w aggressive valuate_path i =
  let nod = t.nodes.(i) in
  
  
  let cops_proximity = compute_cops_proximity t w 4 i in
  
  let arrested_risk = 
    let risk = ref 0 in
    for cop = 0 to t.ncops - 1 do
      match w.cops_ptypes.(cop) with
        Cop_car ->
          if t.car_dist.(w.cops_nodes.(cop)).(i) <= 1 then begin
(*              Printf.fprintf stderr "Arrested risk by car cop in %d\n" i; *)
              incr risk
            end
      | Cop_foot ->
          if t.foot_dist.(w.cops_nodes.(cop)).(i) <= 1 then begin
(*              Printf.fprintf stderr "Arrested risk by foot cop in %d\n" i; *)
              incr risk
            end
      | Robber -> assert false
    done;
    !risk
  in
  
  let smelled_risk = 
    let risk = ref (2 * arrested_risk) in
    for cop = 0 to t.ncops - 1 do
      match w.cops_ptypes.(cop) with
        Cop_car -> ()
      | Cop_foot ->
          if t.foot_dist.(w.cops_nodes.(cop)).(i) = 2 then
            incr risk 
      | Robber -> assert false
    done;
(*    if !risk > 0 then 1 else 0 *) !risk
  in
  
  let bank_distance = Robber_misc.bank_distance t w i in
  
  let node_degree = List.length t.foot_edges.(i) in
  
  let copsum = Robber_misc.sum_to_cops1 t w i in
  let copsum = 
    if copsum < copsum_threshold then - copsum else - copsum_threshold
  in
  let copsum = 0 in
  let paths = valuate_path i in
  
  
  let is_in_culdesac =  nod.node_danger in 

(* A bad hack to rob a bank at the last moment 
  let arrested_risk =
    if !copdists.(t.robber_node) - bank_distance >= (200 - w.state_num) / 2 then
      -1
    else
    if  !copdists.(t.robber_node) - bank_distance > 1 + t.robberies &&
      !copsums.(t.robber_node) > 25
      then -1
    else arrested_risk
  in
*)
  match aggressive with
    1 ->
      ( arrested_risk, smelled_risk, is_in_culdesac, bank_distance, -paths, 
        copsum,
        cops_proximity/3, node_degree, 
        (if i = t.robber_node then 1 else 0)
      )
  | 2 ->
      ( arrested_risk, smelled_risk, bank_distance, -paths, 
        is_in_culdesac, copsum,
        cops_proximity/3, node_degree, 
        (if i = t.robber_node then 1 else 0)
      )
  | _ ->
      ( arrested_risk, is_in_culdesac, smelled_risk, -paths, 
        bank_distance, copsum,
        cops_proximity/3, node_degree, 
        (if i = t.robber_node then 1 else 0)
      )
      
let inform t w = 
  
  let valuate_path = 
    Robber_misc.sum_to_cops_from t w  [t.robber_node] in
  
  let valuate_path = 
    Robber_misc.path_from t w 
      (fun len i -> 
        
        t.nodes.(i).node_copsum
(*        let dist = !copdists.(i) in
(5 - len) * (max (dist - len) 0) * !copsums.(i) *)
    
    ) (fun len i -> 
        
        (*
        2 * copsums.(i)
        let dist = !copdists.(i) in
        dist * !copsums.(i)
*)
        3 * t.nodes.(i).node_copsum - 10 * t.nodes.(i).node_danger
    ) 7 [t.robber_node]  in
  
  let neighbours = t.robber_node :: t.foot_edges.(t.robber_node) in 
  let cops_proximity5 = compute_cops_proximity t w 6 t.robber_node in
  let aggressive = 
    if cops_proximity5 < 1 then begin
        Printf.fprintf stderr "----------   AWAY FROM COPS ------------\n";
        2
      end else
    if cops_proximity5 < 2 then begin
        Printf.fprintf stderr "----------   FAR FROM COPS ------------\n";
        1
      end else
(*    if cops_proximity5 = 5 then begin
        Printf.fprintf stderr "----------   NO HOPE ------------\n";
        1
      end else  *)
      0

  in
  let neighbours = List.map (fun i ->
        compute_node_value t w aggressive valuate_path i , i        
    ) neighbours in
  
  let neighbours = List.sort compare neighbours in
  let next_node = snd (List.hd neighbours) in
  t.proposed_robber_node <- next_node;
  
  Printf.fprintf stderr "\nCurrent position: %d\n" t.robber_node;
  Printf.fprintf stderr "Move sorted:\n";
  List.iter (fun ( (a,b,c,d,e,f,g,h,i), num) ->
      Printf.fprintf stderr "    %d -> %d, %d, %d, %d, %d, %d, %d, %d, %d \n"
        num  a b c d e f g h i
  ) neighbours;
  
  Printf.fprintf stderr "Next position: %d\n" t.proposed_robber_node;
  flush stderr;
  []
  
let _ = 
  register_robber "robber7" 
    "revision of robber5"  
    {
    Robber1.basic_robber with
    robber_inform = inform;
    robber_sumcops = Robber_misc.sum_to_cops1;
  }
  