(* $Id: robber8.ml,v 1.4.2.2 2005/07/10 12:07:17 lefessan Exp $ *)


(* 

TODO list:

* Si on arrive a une banque en n mod 8 <> 0, on peut attendre qques tours
pour rentrer dans la banque uniquement en n mod 8 = 0.
* Si un flic reste a cote d'une banque, on ne s'interesse pas a cette banque.
* Ne pas aller systematiquement vers la banque la plus proche.
* Fixer les destinations une fois pour toute
*)

open Args
open Message
open Robber_types
open Robber_topology
open Robber
  
let copsum_threshold = 35  

type aggressivity =
  CloseToCops
| FarFromCops
| AwayFromCops
  
let compute_cops_proximity t w dist_threshold i = 
  let sum = ref 0 in
  for cop = 0 to t.ncops - 1 do
    match w.cops_ptypes.(cop) with
      Cop_car ->
        if t.car_dist.(w.cops_nodes.(cop)).(i) < dist_threshold then begin
(*            Printf.fprintf stderr "Car cop closed to %d : %d\n" 
              i t.car_dist.(w.cops_nodes.(cop)).(i); *)
            incr sum
          end
    | Cop_foot ->
        if t.foot_dist.(w.cops_nodes.(cop)).(i) < dist_threshold then begin
(*            Printf.fprintf stderr "Foot cop closed to %d : %d\n" 
              i t.foot_dist.(w.cops_nodes.(cop)).(i); *)
            incr sum
          end
    | Robber -> assert false
  done;
  !sum
  
  
let compute_node_value t w aggressive valuate_path i =
  let nod = t.nodes.(i) in
  
  
  let cops_proximity = compute_cops_proximity t w 4 i in
  
  let arrested_risk = 
    let risk = ref 0 in
    for cop = 0 to t.ncops - 1 do
      match w.cops_ptypes.(cop) with
        Cop_car ->
          if t.car_dist.(w.cops_nodes.(cop)).(i) <= 1 then begin
(*              Printf.fprintf stderr "Arrested risk by car cop in %d\n" i; *)
              incr risk
            end
      | Cop_foot ->
          if t.foot_dist.(w.cops_nodes.(cop)).(i) <= 1 then begin
(*              Printf.fprintf stderr "Arrested risk by foot cop in %d\n" i; *)
              incr risk
            end
      | Robber -> assert false
    done;
    !risk
  in
  
  let smelled_risk = 
    let risk = ref (2 * arrested_risk) in
    for cop = 0 to t.ncops - 1 do
      match w.cops_ptypes.(cop) with
        Cop_car -> ()
      | Cop_foot ->
          if t.foot_dist.(w.cops_nodes.(cop)).(i) = 2 then
            incr risk 
      | Robber -> assert false
    done;
    if !risk > 0 then 1 else 0
  in
  
  let bank_distance = t.nodes.(i).node_bankdist in
  
  let node_degree = List.length t.foot_edges.(i) in
  
  let paths = valuate_path i in

  
  let is_in_culdesac = nod.node_danger in 

  let robbernode = t.nodes.(t.robber_node) in
  let copsum = robbernode.node_copsum in  
  let arrested_risk =
    if robbernode.node_copdist - bank_distance >= (200 - w.state_num) / 2 then
      -1
    else arrested_risk
  in

  match aggressive with
    FarFromCops -> (* FAR FROM COPS *)
      ( arrested_risk, smelled_risk, is_in_culdesac, bank_distance, -paths, 
        copsum,
        cops_proximity/3, node_degree, 
        (if i = t.robber_node then 1 else 0)
      )
  | AwayFromCops -> (* AWAY FROM COPS *)
      ( arrested_risk, smelled_risk, bank_distance, -paths, 
        is_in_culdesac, copsum,
        cops_proximity/3, node_degree, 
        (if i = t.robber_node then 1 else 0)
      )
  | CloseToCops -> (* NORMAL *)
      ( arrested_risk, smelled_risk, -paths, 
        is_in_culdesac, bank_distance, copsum,
        cops_proximity/3, node_degree, 
        (if i = t.robber_node then 1 else 0)
      )

      
      
let inform t w = 
  
  let valuate_path = 
    Robber_misc.sum_to_cops_from t w  [t.robber_node] in
  
  let valuate_path = 
    Robber_misc.path_from t w 
      (fun len i -> 
        let node = t.nodes.(i) in
        let total = (1+len) * node.node_copsum in
        let total = total - 10 * node.node_danger in
        let total = total + 3 * node.node_density in
        let total = match node.node_tag with
            Bank _ -> total - 200 / (1+node.node_copdist)
          | _ -> total
        in
        let total = if i = t.last_robber_node then
            total - 50
          else total in 
        total
          
(*        let dist = !copdists.(i) in
(5 - len) * (max (dist - len) 0) * !copsums.(i) *)
    
    ) (fun len i -> 
        let node = t.nodes.(i) in
        let total = (2*len) * node.node_copsum in
        let total = total - 10 * node.node_danger in
        let total = total + 3 * node.node_density in
        let total = match node.node_tag with
            Bank _ -> total - 200 / (1+node.node_copdist)
          | _ -> total
        in
        let total = if i = t.last_robber_node then
            total - 50
          else total in 
        total
    ) 7 [t.robber_node]  in
  
  let neighbours = t.foot_edges.(t.robber_node) in 
  
  let (all_cops, foot_cops, car_cops) = cops_proximities t w t.robber_node in
  let cops_proximity = compute_cops_proximity t w 5 t.robber_node in
  let aggressive = 
    if all_cops.(5) < 1 then begin
        Printf.fprintf stderr "----------   AWAY FROM ALL COPS ------------\n";
        AwayFromCops
      end else
    if foot_cops.(7) < 1 && car_cops.(2) < 1 then begin
        Printf.fprintf stderr "----------   AWAY FROM CAR COPS ------------\n";
        AwayFromCops
      end else
    if all_cops.(5) < 2 then begin
        Printf.fprintf stderr "----------   FAR FROM COPS ------------\n";
        FarFromCops
      end else begin
(*    if cops_proximity5 = 5 then begin
        Printf.fprintf stderr "----------   NO HOPE ------------\n";
        1
end else  *)
        Printf.fprintf stderr "----------   CLOSE TO COPS ------------\n";
      CloseToCops
      end
  in
  let list = ref [] in
  for i = 0 to t.ncops - 1 do
    match w.cops_ptypes.(i) with
      Cop_car ->
        list := (
          t.car_dist.(w.cops_nodes.(i)).(t.robber_node), 
          "car",
          t.cops_names.(i), 
          w.cops_nodes.(i)
       ) :: !list
    | Cop_foot ->
        list := (t.foot_dist.(w.cops_nodes.(i)).(t.robber_node), "foot",
          t.cops_names.(i),
          w.cops_nodes.(i)          
        ) :: !list
    | _ -> ()
  done;
  let list = List.sort compare !list in
  Printf.fprintf stderr "COPS: \n";
  List.iter (fun (dist, ptype, name, pos) ->
      Printf.fprintf stderr " %d (%s at %d: %s)" dist name pos  ptype
  ) list;
  Printf.fprintf stderr "\n";
  
  let neighbours = List.map (fun i ->
        compute_node_value t w aggressive valuate_path i , i        
    ) neighbours in
  
  let neighbours = List.sort compare neighbours in
  let next_node = snd (List.hd neighbours) in
  t.proposed_robber_node <- next_node;
  
  Printf.fprintf stderr "\nCurrent position: %d\n" t.robber_node;
  Printf.fprintf stderr "Move sorted:\n";
  List.iter (fun ( (a,b,c,d,e,f,g,h,i), num) ->
      Printf.fprintf stderr "    %d -> %d, %d, %d, %d, %d, %d, %d, %d, %d \n"
        num  a b c d e f g h i
  ) neighbours;
  
  Printf.fprintf stderr "Next position: %d\n" t.proposed_robber_node;
  flush stderr;
  []
  
let _ = 
  register_robber "robber8" 
    "revision of robber7"
    {
    Robber1.basic_robber with
    robber_inform = inform;
    robber_sumcops = Robber_misc.sum_to_cops2;
  }
  