(* $Id: robber_misc.ml,v 1.6.2.2 2005/07/10 12:07:17 lefessan Exp $ *)

open Message
open Robber_types
open Robber_topology

  
  
  
  
let sum_to_cops1 t w i =
  let sum = ref 0 in
  for cop = 0 to t.ncops - 1 do
    if not w.cops_bribed.(cop) then
    let dist = match w.cops_ptypes.(cop) with
        Cop_car ->
          let dist = t.car_dist.(w.cops_nodes.(cop)).(i) in          
          (match dist with
              0 | 1 -> 0
            | 2 -> 1
            | 3 | 4 -> 3
            | 5 | 6 -> 5
            | _ -> 8) 
      | Cop_foot ->
          let dist = t.foot_dist.(w.cops_nodes.(cop)).(i) in
          (match dist with
              0 -> 0
            | 1 | 2 -> 0
            | 3 -> 1
            | 4 -> 3
            | 5 | 6 -> 5
            | _ -> 8)
      | Robber -> 0
    in
    sum := !sum +  dist
  done;
  !sum
  
let sum_to_cops2 t w i =
  let sum = ref 0 in
  for cop = 0 to t.ncops - 1 do
    if not w.cops_bribed.(cop) then
    let dist = match w.cops_ptypes.(cop) with
        Cop_car ->
          let dist = t.car_dist.(w.cops_nodes.(cop)).(i) in          
          2 * (min dist 5)
      | Cop_foot ->
          let dist = t.foot_dist.(w.cops_nodes.(cop)).(i) in
          min dist 10
      | Robber -> 0
    in
    sum := !sum +  dist
  done;
  !sum
  
let sum_to_cops3 t w i =
  let sum = ref 0 in
  for cop = 0 to t.ncops - 1 do
    if not w.cops_bribed.(cop) then
    let dist = match w.cops_ptypes.(cop) with
        Cop_car ->
          let dist = t.car_dist.(w.cops_nodes.(cop)).(i) in          
          let dist = if dist < 5 then dist else 5 in
          3 * dist
      | Cop_foot ->
          let dist = t.foot_dist.(w.cops_nodes.(cop)).(i) in
          let dist = if dist < 8 then dist else 10 in
          2 * dist
      | Robber -> 0
    in
    sum := !sum +  dist
  done;
  !sum

let dist_to_cops t w i =
  let sum = ref max_int in
  for cop = 0 to t.ncops - 1 do
    if not w.cops_bribed.(cop) then
    let dist = match w.cops_ptypes.(cop) with
        Cop_car ->
          t.car_dist.(w.cops_nodes.(cop)).(i) 
      | Cop_foot ->
          t.foot_dist.(w.cops_nodes.(cop)).(i)
      | Robber -> 0
    in
    sum := min !sum dist
  done;
  !sum

  
let compute_cops_proximity t w dist_threshold i = 
  let sum = ref 0 in
  for cop = 0 to t.ncops - 1 do
    if not w.cops_bribed.(cop) then
    match w.cops_ptypes.(cop) with
      Cop_car ->
        if t.car_dist.(w.cops_nodes.(cop)).(i) < dist_threshold then begin
(*            Printf.fprintf stderr "Car cop closed to %d : %d\n" 
              i t.car_dist.(w.cops_nodes.(cop)).(i); *)
            incr sum
          end
    | Cop_foot ->
        if t.foot_dist.(w.cops_nodes.(cop)).(i) < dist_threshold then begin
(*            Printf.fprintf stderr "Foot cop closed to %d : %d\n"
              i t.foot_dist.(w.cops_nodes.(cop)).(i); *)
            incr sum
          end
    | Robber -> assert false
  done;
  !sum
  
let bank_distance t w i =
  let dist = ref 10000 in
  for bank = 0 to t.nbanks - 1 do
    let bank = t.banks.(bank) in
    let bank_node = bank.bank_node in
    if bank_node <> t.last_robbery then
      let cops_proximity = compute_cops_proximity t w 3 bank_node in
      dist := min !dist (
        if cops_proximity < 2 then
          t.foot_dist.(i).(bank_node)
        else
          6 * cops_proximity +  t.foot_dist.(i).(bank_node)
      )
  done;
  !dist
  
let update_nodes t w r =
  for i = 0 to !nnodes - 1 do
    let node = t.nodes.(i) in
    node.node_copsum <- r.robber_sumcops t w i;
    node.node_bankdist <- bank_distance t w i;
    node.node_copdist <- dist_to_cops t w i;
  done
  
let init_dists_to_cops t w =
  let sums = Array.create !nnodes 0 in
  for i = 0 to !nnodes - 1 do
    sums.(i) <- dist_to_cops t w i
  done;
  sums

let sum_to_cops_from t w avoid i =
  
  let path_len = 6 in
  let rec iter1 i avoid len sum maxi =
    if len = path_len then 
      if sum > maxi then sum else maxi else
    let edges = t.foot_edges.(i) in
    iter2 i avoid edges len sum maxi
    
  and iter2 i avoid edges len sum maxi =
    match edges with
      [] -> maxi
    | edge :: tail ->
        let maxi = 
          if List.mem edge avoid then maxi else
            iter1 edge (i :: avoid) (len+1) 
            (sum + t.nodes.(edge).node_copsum * (10 + len) / 10 ) maxi in
        iter2 i avoid tail len sum maxi
  in
  iter1 i avoid 1 0 0
  
  
let path_from t w valuate_step valuate_end path_len avoid i =
  let rec iter1 i avoid len sum maxi =
    if len = path_len then 
      let sum = sum + valuate_end len i in
      if sum > maxi then sum else maxi else
    let edges = t.foot_edges.(i) in
    iter2 i avoid edges len 
      (sum + valuate_step len i) maxi
    
  and iter2 i avoid edges len sum maxi =
    match edges with
      [] -> maxi
    | edge :: tail ->
        let maxi = 
          if List.mem edge avoid then maxi else
            iter1 edge (i :: avoid) (len+1) 
            sum maxi in
        iter2 i avoid tail len sum maxi
  in
  iter1 i avoid 1 0 0
  
