(* $Id: robber_planning.ml,v 1.1.2.3 2005/07/10 12:07:17 lefessan Exp $ *)

open Args
open Message
open Robber_types
open Robber_topology
open Robber
  
let votes = ref []

let planning t s =  
  Printf.fprintf stderr "Robber_planning.planning...\n"; 
  let plans = ref [
      { 
        plan_bot = t.our_name;
        plan_loc = loc_of_node t t.proposed_robber_node;
        plan_world = s.state_num;
        plan_ptype = Robber;
      }
      ] in
  for c = 0 to t.ncops - 1 do
    if s.cops_bribed.(c) then begin
        let dists, edges = match s.cops_ptypes.(c) with
            Cop_car -> t.car_dist, t.car_edges
          | Cop_foot -> t.foot_dist, t.foot_edges
          | _ -> assert false
        in
        
        let node = s.cops_nodes.(c) in
        let dests = node :: edges.(node) in
        let list = List.sort compare 
          (List.map (fun n ->
              - dists.(n).(t.proposed_robber_node), n
            ) dests)
        in
        let rec iter best list =
          match list with
            [] -> best
          | (dist, n) :: tail ->
              let best =
                if dist = -2 then n else best
              in
              iter best tail
        in
        let best = iter node list in
        Printf.fprintf stderr " move %s to %d\n" t.cops_names.(c) best;
        plans := {
          plan_bot = t.cops_names.(c);
          plan_loc = loc_of_node t best;
          plan_ptype = s.cops_ptypes.(c);
          plan_world = s.state_num;
        } :: !plans
        
      end
  done;
  !plans
  
let voting t list =
  Printf.fprintf stderr "Robber_planning.voting...\n"; 
  snd (
    List.split (
      List.sort compare 
        (List.map (fun (c, _) ->
            let nvotes = 
              try
                ! (List.assoc c !votes)
              with Not_found -> 
                  let nvotes =
                    if c = t.our_name then -10000 else 0 in
                  votes := (c, ref nvotes) :: !votes;
                  nvotes
            in
            nvotes, c
        )
        list
      )
    )
  )
  
let count vote_res =
  Printf.fprintf stderr "Robber_planning.count...\n"; 
  match vote_res with
  | Nowinner -> ()
  | Winner w ->
      try
        incr (List.assoc w (!votes))
      with Not_found ->  (* This should not occur, but still... *)
(*          voting_scores := (w, ref 1) :: !voting_scores *)
          assert false
          