(* $Id: robber_topology.ml,v 1.2.2.5 2005/07/10 12:07:17 lefessan Exp $ *)

open Args
open Message
open Robber_types
  
let num_of_cop t name =
  Hashtbl.find t.cop_by_name name

let loc_of_node t i = t.nodes.(i).node_loc
let node_of_loc t loc = Hashtbl.find t.node_by_loc loc
  
let dummy_node = {
    node_num = 0;
    node_loc = "";
    node_tag = Ordinary;
    node_x = 0;
    node_y = 0;
    node_danger = 0;
    node_copsum = 0;
    node_bankdist = 0;
    node_copdist = 0;
    node_density = 0;
  }

let map_create () = {
    nbanks = 0;
    ncops = 0;
    
    our_name = "";
    robber_name = "";
    cops_names = [||];
    
    nodes = [||];
    node_by_loc = Hashtbl.create 133;
    cop_by_name = Hashtbl.create 133;
    
    initial_cop_node = 0 ;
    cop_node = 0 ;
    cop_ptype = Cop_foot ;

    last_robber_node = 0;
    proposed_robber_node = 0;
    proposed_robber_bribe = Nothing;
    initial_robber_node = 0 ;
    robber_node = 0 ;
    
    banks_nodes = Intset.empty;
    banks = [||] ;
    car_edges = [||] ;
    foot_edges = [||] ;
    
    car_adj = [||];
    car_dist = [||];
    foot_adj = [||];
    foot_dist = [||];

    last_robbery_time = 0;
    last_robbery = -1;
    robberies = 0;
    stolen = 0;
    
    states = [];
  }

  
let build_adj edges =
  let adj = Array.create_matrix !nnodes !nnodes 0 in
  Array.iteri (fun src list ->
      List.iter (fun dst ->
          adj.(src).(dst) <- 1
      ) list
  ) edges;
  adj

let floyd_warshal way adj =
  let nnodes = !nnodes in
  let dist = Array.create_matrix nnodes nnodes 10000 in
  
  let dist = dist in
  for i = 0 to nnodes - 1 do
    dist.(i).(i) <- 0;
  done;
  for src = 0 to nnodes - 1 do
    let adj = adj.(src) in
    let dist = dist.(src) in
    for dst = 0 to nnodes - 1  do
      if adj.(dst) = 1 then
        dist.(dst) <- 1
    done
  done;
  for k = 0 to nnodes - 1 do
    let distk = dist.(k) in
    for i = 0 to nnodes - 1 do
      let disti = dist.(i) in
      for j = 0 to nnodes - 1 do
        let d = disti.(k) + distk.(j) in
        if d < disti.(j) then disti.(j) <- d
      done
    done
  done;
  let oc = open_out (
      Printf.sprintf "/tmp/distances-%s.txt" way) in
  for i = 0 to nnodes - 1 do
    for j = 0 to nnodes - 1 do
      Printf.fprintf oc "%d -> %d : %d\n" i j dist.(i).(j)
    done
  done;
  close_out oc;
  dist
  
      
let density t i =
  let set = ref Intset.empty in
  let rec iter len n =
    if len > 0 then 
      let edges = t.foot_edges.(n) in
      List.iter (fun n ->
          set := Intset.add n !set;
          iter (len-1) n
      ) edges
  in
  iter 3 i;
  Intset.cardinal !set

let find_bad_nodes t =

  for i = 0 to !nnodes - 1 do
    
    let marks = Array.create !nnodes false in
    
    let rec dfs i =
      if not marks.(i) then begin
          marks.(i) <- true;
          List.iter (fun dst ->
              dfs dst
          ) t.foot_edges.(i)
          
        end
    in
    marks.(i) <- true;
    let first = if i = 0 then 1 else 0 in
    dfs first;
    
    for j = 0 to !nnodes - 1 do
      if not marks.(j) then 
        t.nodes.(j).node_danger <- 10
    done;
    
    t.nodes.(i).node_density <- density t i;
  done

let node_of_nod nod =
  {
    node_num = nod.nod_num;
    node_loc = nod.nod_loc;
    node_tag = Ordinary;
    node_x = nod.nod_x;
    node_y = nod.nod_y;
    node_danger = 0;
    node_copsum = 0;
    node_bankdist = 0;
    node_copdist = 0;
    node_density = 0;
  }

  
let map_init t wsk = 
  let begin_time = Unix.gettimeofday () in
  t.our_name <- wsk.wsk_name;
  t.robber_name <- wsk.wsk_robber;
  t.cops_names <- Array.of_list wsk.wsk_cops;
  t.ncops <- Array.length t.cops_names;
  for i = 0 to t.ncops - 1 do
    Hashtbl.add t.cop_by_name t.cops_names.(i) i
  done;
  t.nodes <- Array.create !nnodes dummy_node;
  let banks = ref [] in
  List.iter (fun nod ->
      let node = node_of_nod nod in
      t.nodes.(node.node_num) <- node;
      Hashtbl.add t.node_by_loc node.node_loc node;
      match nod.nod_tag  with
        Message.Ordinary -> ()
      | Message.Hq -> 
          node.node_tag <- Hq;
          t.cop_node <- node.node_num;
          t.initial_cop_node <- node.node_num
      | Message.Bank -> 
          let bank = {
              bank_node = node.node_num;
              bank_num = t.nbanks;
              bank_value = 0;
            } in
          t.nbanks <- t.nbanks + 1;
          banks := bank :: !banks;
          node.node_tag <- Bank bank
      | Message.Robber_start ->
          t.robber_node <- node.node_num;
          t.initial_robber_node <- node.node_num;
  ) wsk.wsk_nods;
  t.banks <- Array.of_list (List.rev !banks);
  
  Printf.fprintf stderr "%d sommets\n" !nnodes; 
  t.car_edges <- Array.create !nnodes [];
  t.foot_edges <- Array.create !nnodes [];
  List.iter (fun edg ->
      let src = node_of_loc t edg.edg_src in
      let dst = node_of_loc t edg.edg_dst in
      t.car_edges.(src.node_num) <- dst.node_num :: t.car_edges.(src.node_num);
      match edg.edg_type with
        Car -> ()
      | Foot ->
          
          let add_edge src dst =
            if not (List.mem dst.node_num t.foot_edges.(src.node_num)) then
              t.foot_edges.(src.node_num) <- 
                dst.node_num :: t.foot_edges.(src.node_num)
          in
          add_edge src dst;
          add_edge dst src;
  
  
  ) wsk.wsk_edgs;
  
  let init_time = Unix.gettimeofday () in
  Printf.fprintf stderr "Init time : %.2f s\n" (init_time -. begin_time);
  let begin_time = init_time in
  
  t.foot_adj <- build_adj t.foot_edges;
  
  let init_time = Unix.gettimeofday () in
  Printf.fprintf stderr "foot_adj time : %.2f s\n" (init_time -. begin_time);
  let begin_time = init_time in
  
  t.foot_dist <- floyd_warshal "foot" t.foot_adj;
  
  let init_time = Unix.gettimeofday () in
  Printf.fprintf stderr "foot_dist time : %.2f s\n" (init_time -. begin_time);
  let begin_time = init_time in
  
  t.car_adj <- build_adj t.car_edges;
  
  let init_time = Unix.gettimeofday () in
  Printf.fprintf stderr "car_adj time : %.2f s\n" (init_time -. begin_time);
  let begin_time = init_time in
  
  t.car_dist <-  floyd_warshal "car" t.car_adj;
  
  let init_time = Unix.gettimeofday () in
  Printf.fprintf stderr "car_dist time : %.2f s\n" (init_time -. begin_time);
  let begin_time = init_time in
  
  find_bad_nodes t;      
  
  let init_time = Unix.gettimeofday () in
  Printf.fprintf stderr "bad nodes time : %.2f s\n" (init_time -. begin_time);
 (* let begin_time = init_time in (* UNUSED *) *)
  ()
  


let cops_proximities t w i =
  let all_cops = Array.create 10 0 in
  let car_cops = Array.create 10 0 in
  let foot_cops = Array.create 10 0 in
  for cop = 0 to t.ncops - 1 do
    match w.cops_ptypes.(cop) with
      Cop_car ->
        let dist = t.car_dist.(w.cops_nodes.(cop)).(i) in
        let dist = if dist > 9 then 9 else dist in
        for i = dist to 9 do
          car_cops.(i) <- 1 + car_cops.(i);
          all_cops.(i) <- 1 + all_cops.(i);
        done;
    | Cop_foot ->
        let dist = t.foot_dist.(w.cops_nodes.(cop)).(i) in
        let dist = if dist > 9 then 9 else dist in
        for i = dist to 9 do
          foot_cops.(i) <- 1 + foot_cops.(i);
          all_cops.(i) <- 1 + all_cops.(i);
        done;
    | Robber -> assert false
  done;
  (all_cops, foot_cops, car_cops)
  
let state_create t r wd = 
  
  
  let cops_nodes = Array.create t.ncops 0 in
  let cops_ptypes = Array.create t.ncops Cop_foot in
  
  
  Printf.fprintf stderr "\nWorld %d\n" wd.wor_wor;
  List.iter (fun bv ->
      let nod = node_of_loc t bv.bv_loc in
      match nod.node_tag with
        Bank bank ->
          bank.bank_value <- bv.bv_value
      | _ -> assert false
  ) wd.wor_bvs;
  Printf.fprintf stderr "\n";
  List.iter (fun pl ->
      match pl.pl_ptype with
      | Cop_foot ->
          let ncop = num_of_cop t pl.pl_bot in
          cops_nodes.(ncop) <- (node_of_loc t pl.pl_loc).node_num;
          cops_ptypes.(ncop) <- Cop_foot;
      | Cop_car ->
          let ncop = num_of_cop t pl.pl_bot in
          cops_nodes.(ncop) <- (node_of_loc t pl.pl_loc).node_num;
          cops_ptypes.(ncop) <- Cop_car;
      | _ -> 
          t.robber_node <- (node_of_loc t pl.pl_loc).node_num
  ) wd.wor_pls;
  let cops_bribed = Array.create t.ncops false in
  List.iter (fun dc ->
      let ncop = num_of_cop t dc in
      cops_bribed.(ncop) <- true
  ) wd.wor_dc;
  
  
  
  let w = {
      state_robber = r;
      state_num = wd.wor_wor;
      cops_nodes = cops_nodes;
      cops_ptypes = cops_ptypes;
      cops_bribed = cops_bribed;
      bank_all_cops = Array.create t.nbanks [||];
      bank_car_cops = Array.create t.nbanks [||];
      bank_foot_cops = Array.create t.nbanks [||];
      state_robbernode = (-1);
      state_possibles = [||];
      state_known = false;
      state_messages = [];
    } in

  for i = 0 to t.nbanks - 1 do
    let (all_cops, foot_cops, car_cops) = 
      cops_proximities t w t.banks.(i).bank_node in
    w.bank_all_cops.(i) <- all_cops;
    w.bank_car_cops.(i) <- car_cops;
    w.bank_foot_cops.(i) <- foot_cops;
  done;
  t.states <- w :: t.states;
  w  
