(* $Id: topology.mli,v 1.25.4.3 2005/07/10 10:11:26 frisch Exp $ *)

open Message;;

module NS : Set.S with type elt = nod

type topo = {
  t_nods : nod array;
  t_nods_alpha : nod array;
  t_neigh_f : int list array;
  t_neigh_c : int list array;
  t_neigh_c_inv : int list array;

  t_dist_f : int array Lazy.t array;
  t_dist_c : int array Lazy.t array;

  t_dist_f_1 : NS.t array;
  t_dist_f_2 : NS.t array;
  t_dist_c_1 : NS.t array;

  t_banks : NS.t;
  all_nodes : NS.t;

  ncops : int;
  cops_name : string array;
  cops_number : (string * int) list;

  initial_robber : nod;
  robber_name : string;

  hq : nod;

  name : string;
}

val nodes_count : topo -> int

val read_skel : world_skeleton -> topo;;

val get_node : topo -> string -> nod
val node_of_player : topo -> pl -> nod

val cop_number : topo -> string -> int
(*
val cop_name : topo -> int -> string
*)


val dump : topo -> unit;;

val dist_f : topo -> nod -> nod -> int
val dist_c : topo -> nod -> nod -> int


val dist_f_1 : topo -> nod -> NS.t
val dist_f_2 : topo -> nod -> NS.t
val dist_c_1 : topo -> nod -> NS.t


(* Le voleur est dans un ensemble au tour n. Ou peut-il etre au tour suivant 
   (ou au tour precedent) ? *)
val extend : topo -> NS.t -> NS.t

(* Le meme pour n tours *)
val extend_n : topo -> int -> NS.t -> NS.t

(* Le voleur tait  la position x au tour n. Ou peut-il etre au tour n' ? *)
val was_at : topo -> nod -> int -> int -> NS.t

(* Un policier sent. Ou peut etre le voleur ? *)
val smell : topo -> nod -> ptype -> int -> NS.t

(* Un joueur (qui peut-etre controle plusieurs policiers) sent. 
   Ou peut etre le voleur ? *)
val player_smell : topo -> (nod * ptype) list -> int -> NS.t


(* Un policier ne sent rien. Ou le voleur ne peut-il *pas* etre ? *)
val nosmell : topo -> nod -> ptype -> NS.t

(* Un joueur (qui peut-etre controle plusieurs policiers) ne sent rien.
   Ou le voleur ne peut-il *pas* etre ? *)
val player_nosmell : topo -> (nod * ptype) list -> NS.t

(* Destinations possible sans changer de mode *)
val possible_dest : topo -> nod -> ptype -> NS.t

(* val possible_dest_robot : topo -> pl -> NS.t *)



(* Distance minimale  un ensemble *)
val dist_to_set : topo -> nod -> NS.t -> int
val dist_to_set_typ : topo -> nod -> NS.t -> ptype -> int

val is_adjacent : topo -> Message.ptype -> nod -> nod -> bool


(* Vrifier que les mouvements de mes policier sont lgaux. Sinon
   rester sur place *)

val can_i_move: topo -> Message.world -> (int * (ptype * nod)) list -> moves
val is_legal: topo -> Message.world -> (int * (ptype * nod)) list -> bool
