(* $Id: xavier_cop.ml,v 1.18.4.5 2005/07/10 13:24:25 frisch Exp $ *)

open Message
open Printf
open Topology

type state = {
  planner: Planner.state;
  voter: Voting.t;
}

let xavier_cop me env w state =
  let () = 
    let info = env.Cops_api.env_info in
    env.Cops_api.env_info <- 
      Info.change_cop info (Planify.change_cops_move_type info);
    (* Changement de moyen de locomotion *) in

  let topo = env.Cops_api.env_topo
  and drawer = env.Cops_api.env_drawer in

  let infos = Cop_informer.compute_inform env.Cops_api.env_info w in

  let informs = Cops_api.cop_inform me env infos in

(*
  let () =
    let now = env.Cops_api.env_info in
    if (now > 1) && ((now - 1) mod 8 = 0) 
    match Cops_api.get_old_info env (Info.current_world now - 2) with
      | Some past ->
	  if Info.may_opt ~past ~now w informs then
	    assert false
      | None -> ()
  in
*)
	  
(*
    eprintf "set sizes = %d %d\n"
      (NS.cardinal (Info.get_possible info))
      (NS.cardinal (Info.get_certain info));
    flush stderr;
*)
    Draw.display_node_sets drawer
      (Info.get_possible env.Cops_api.env_info)
      NS.empty (*(Info.get_certain info)*);

  let (scores, plan) = Planner.make_plan state.planner env.Cops_api.env_info
    (Info.get_certain env.Cops_api.env_info)
    (Info.get_possible env.Cops_api.env_info) in
  let _plans = Cops_api.cop_plan me plan in

  let vote = Voting.vote state.voter topo.name (List.map fst _plans) in
  let vote_res = Cops_api.cop_vote me vote in
  Voting.count state.voter vote_res;

  let winning_plan =
    match vote_res with
    | Nowinner -> []
    | Winner lucky_guy ->
        try List.assoc lucky_guy _plans with Not_found -> [] in

  let my_move = 
    Planner.decide_move state.planner env.Cops_api.env_info scores winning_plan in

  Straight, my_move, Info.get_ripoux env.Cops_api.env_info,
  state

let () =
  Cops_api.register_cop "xavier_cop"
    { Cops_api.cops_init = (fun t env -> 
        { voter = Voting.init env.Cops_api.env_wsk;
          planner = Planner.init env.Cops_api.env_topo env.Cops_api.env_drawer });
      Cops_api.cops_clean_job  = xavier_cop;
      Cops_api.cops_dirty_job = (fun _ _ _ _ -> assert false);
      Cops_api.cops_dirty_r_job = (fun _ _ _ _ -> assert false);
      Cops_api.cops_ptype = Cop_foot;
    }
